import asyncio
from datetime import datetime

from apscheduler.events import (
    EVENT_JOB_ERROR,
    EVENT_JOB_EXECUTED,
    EVENT_JOB_MISSED,
    EVENT_SCHEDULER_STARTED,
)
from bilireq.grpc.dynamic import grpc_get_user_dynamics
from bilireq.grpc.protos.bilibili.app.dynamic.v2.dynamic_pb2 import DynamicType
from grpc import StatusCode
from grpc.aio import AioRpcError
from nonebot.adapters.onebot.v11.message import MessageSegment
from nonebot.log import logger

from ...config import plugin_config
from ...database import DB as db
from ...database import dynamic_offset as offset
from ...utils import get_dynamic_screenshot, safe_send, scheduler


async def dy_sched():
    """动态推送"""
    uid = await db.next_uid("dynamic")
    if not uid:
        # 没有订阅先暂停一秒再跳过，不然会导致 CPU 占用过高
        await asyncio.sleep(1)
        return
    user = await db.get_user(uid=uid)
    assert user is not None
    name = user.name

    logger.debug(f"爬取动态 {name}（{uid}）")
    try:
        # 获取 UP 最新动态列表
        dynamics = (
            await grpc_get_user_dynamics(
                uid,
                timeout=plugin_config.haruka_dynamic_timeout,
                proxy=plugin_config.haruka_proxy,
            )
        ).list
    except AioRpcError as e:
        if e.code() == StatusCode.DEADLINE_EXCEEDED:
            logger.error(f"爬取动态超时，将在下个轮询中重试：{e.code()} {e.details()}")
            return
        raise

    if not dynamics:  # 没发过动态
        if uid in offset and offset[uid] == -1:  # 不记录会导致第一次发动态不推送
            offset[uid] = 0
        return
    # 更新昵称
    name = dynamics[0].modules[0].module_author.author.name

    if uid not in offset:  # 已删除
        return
    elif offset[uid] == -1:  # 第一次爬取
        if len(dynamics) == 1:  # 只有一条动态
            offset[uid] = int(dynamics[0].extend.dyn_id_str)
        else:  # 第一个可能是置顶动态，但置顶也可能是最新一条，所以取前两条的最大值
            offset[uid] = max(
                int(dynamics[0].extend.dyn_id_str), int(dynamics[1].extend.dyn_id_str)
            )
        return

    dynamic = None
    for dynamic in sorted(dynamics, key=lambda x: int(x.extend.dyn_id_str)):  # 动态从旧到新排列
        dynamic_id = int(dynamic.extend.dyn_id_str)
        if dynamic_id > offset[uid]:
            logger.info(f"检测到新动态（{dynamic_id}）：{name}（{uid}）")
            image, err = await get_dynamic_screenshot(dynamic_id)
            url = f"https://t.bilibili.com/{dynamic_id}"
            if image is None:
                logger.debug(f"动态不存在，已跳过：{url}")
                return
            elif dynamic.card_type in [
                DynamicType.live_rcmd,
                DynamicType.live,
                DynamicType.ad,
                DynamicType.banner,
            ]:
                logger.debug(f"无需推送的动态 {dynamic.card_type}，已跳过：{url}")
                offset[uid] = dynamic_id
                return

            type_msg = {
                0: "发布了新动态",
                DynamicType.forward: "转发了一条动态",
                DynamicType.word: "发布了新文字动态",
                DynamicType.draw: "发布了新图文动态",
                DynamicType.av: "发布了新投稿",
                DynamicType.article: "发布了新专栏",
                DynamicType.music: "发布了新音频",
            }
            message = (
                f"{name} {type_msg.get(dynamic.card_type, type_msg[0])}：\n"
                f"{f'动态图片可能截图异常：{err}' if err else ''}\n"
                f"{MessageSegment.image(image)}\n{url}"
            )

            push_list = await db.get_push_list(uid, "dynamic")
            for sets in push_list:
                await safe_send(
                    bot_id=sets.bot_id,
                    send_type=sets.type,
                    type_id=sets.type_id,
                    message=message,
                    at=bool(sets.at) and plugin_config.haruka_dynamic_at,
                )

            offset[uid] = dynamic_id

    if dynamic:
        await db.update_user(uid, name)


def dynamic_lisener(event):
    if hasattr(event, "job_id") and event.job_id != "dynamic_sched":
        return
    job = scheduler.get_job("dynamic_sched")
    if not job:
        scheduler.add_job(
            dy_sched, id="dynamic_sched", next_run_time=datetime.now(scheduler.timezone)
        )


if plugin_config.haruka_dynamic_interval == 0:
    scheduler.add_listener(
        dynamic_lisener,
        EVENT_JOB_EXECUTED | EVENT_JOB_ERROR | EVENT_JOB_MISSED | EVENT_SCHEDULER_STARTED,
    )
else:
    scheduler.add_job(
        dy_sched,
        "interval",
        seconds=plugin_config.haruka_dynamic_interval,
        id="dynamic_sched",
    )



from typing import List, Optional

from loguru import logger
from nonebot import get_driver
from pydantic import BaseSettings, validator
from pydantic.fields import ModelField


# 其他地方出现的类似 from .. import config，均是从 __init__.py 导入的 Config 实例
class Config(BaseSettings):
    fastapi_reload: bool = False
    haruka_dir: Optional[str] = None
    haruka_to_me: bool = True
    haruka_live_off_notify: bool = False
    haruka_proxy: Optional[str] = None
    haruka_interval: int = 10
    haruka_live_interval: int = haruka_interval
    haruka_dynamic_interval: int = 0
    haruka_dynamic_at: bool = False
    haruka_screenshot_style: str = "mobile"
    haruka_captcha_address: str = "https://captcha-cd.ngworks.cn"
    haruka_browser_ua: Optional[str] = None
    haruka_dynamic_timeout: int = 30
    haruka_dynamic_font_source: str = "system"
    haruka_dynamic_font: Optional[str] = "Noto Sans CJK SC"
    haruka_dynamic_big_image: bool = False
    haruka_command_prefix: str = ""
    # 频道管理员身份组
    haruka_guild_admin_roles: List[str] = ["频道主", "超级管理员"]

    @validator("haruka_interval", "haruka_live_interval", "haruka_dynamic_interval")
    def non_negative(cls, v: int, field: ModelField):
        """定时器为负返回默认值"""
        return field.default if v < 1 else v

    @validator("haruka_screenshot_style")
    def screenshot_style(cls, v: str):
        if v != "mobile":
            logger.warning("截图样式目前只支持 mobile，pc 样式现已被弃用")
        return "mobile"

    class Config:
        extra = "ignore"


global_config = get_driver().config
plugin_config = Config.parse_obj(global_config)



import contextlib
from typing import Optional, List

import httpx
from nonebot.log import logger
from playwright._impl._api_structures import Position
from playwright.async_api import Page, Response
from pydantic import BaseModel
from yarl import URL

from ..config import plugin_config


class CaptchaData(BaseModel):
    captcha_id: str
    points: List[List[int]]
    rectangles: List[List[int]]
    yolo_data: List[List[int]]
    time: int


class CaptchaResponse(BaseModel):
    code: int
    message: str
    data: Optional[CaptchaData]


async def resolve_captcha(url: str, page: Page) -> Page:
    captcha_image_body = ""
    last_captcha_id = ""
    captcha_result = None

    async def captcha_image_url_callback(response: Response):
        nonlocal captcha_image_body
        logger.debug(f"[Captcha] Get captcha image url: {response.url}")
        captcha_image_body = await response.body()

    async def captcha_result_callback(response: Response):
        nonlocal captcha_result, last_captcha_id
        logger.debug(f"[Captcha] Get captcha result: {response.url}")
        captcha_resp = await response.text()
        logger.debug(f"[Captcha] Result: {captcha_resp}")
        if '"result": "success"' in captcha_resp:
            logger.success("[Captcha] 验证码 Callback 验证成功")
            captcha_result = True
        elif '"result": "click"' in captcha_resp:
            pass
        else:
            if last_captcha_id:
                logger.warning(f"[Captcha] 验证码 Callback 验证失败，正在上报：{last_captcha_id}")
                async with httpx.AsyncClient() as client:
                    await client.post(
                        f"{captcha_baseurl}/report",
                        json={"captcha_id": last_captcha_id},
                    )
                last_captcha_id = ""
            captcha_result = False

    captcha_address = URL(plugin_config.haruka_captcha_address)
    page.on(
        "response",
        lambda response: captcha_image_url_callback(response)
        if response.url.startswith("https://static.geetest.com/captcha_v3/")
        else None,
    )
    page.on(
        "response",
        lambda response: captcha_result_callback(response)
        if response.url.startswith("https://api.geetest.com/ajax.php")
        else None,
    )

    with contextlib.suppress(TimeoutError):
        await page.goto(
            url,
            wait_until="networkidle",
            timeout=plugin_config.haruka_dynamic_timeout * 1000,
        )

    captcha_baseurl = f"{captcha_address.scheme}://{captcha_address.host}:{captcha_address.port}/captcha/select"
    while captcha_image_body or captcha_result is False:
        logger.warning("[Captcha] 需要人机验证，正在尝试自动解决验证码")
        captcha_image = await page.query_selector(".geetest_item_img")
        assert captcha_image
        captcha_size = await captcha_image.bounding_box()
        assert captcha_size
        origin_image_size = 344, 384

        async with httpx.AsyncClient() as client:
            captcha_req = await client.post(
                f"{captcha_baseurl}/bytes",
                timeout=10,
                files={"img_file": captcha_image_body},
            )
            captcha_req = CaptchaResponse(**captcha_req.json())
            logger.debug(f"[Captcha] Get Resolve Result: {captcha_req}")
            assert captcha_req.data
            last_captcha_id = captcha_req.data.captcha_id
        if captcha_req.data:
            click_points: List[List[int]] = captcha_req.data.points
            logger.warning(f"[Captcha] 识别到 {len(click_points)} 个坐标，正在点击")
            # 根据原图大小和截图大小计算缩放比例，然后计算出正确的需要点击的位置
            for point in click_points:
                real_click_points = {
                    "x": point[0] * captcha_size["width"] / origin_image_size[0],
                    "y": point[1] * captcha_size["height"] / origin_image_size[1],
                }
                await captcha_image.click(position=Position(**real_click_points))
                await page.wait_for_timeout(800)
            await page.click("text=确认")
            geetest_up = await page.wait_for_selector(".geetest_up", state="visible")
            if not geetest_up:
                logger.warning("[Captcha] 未检测到验证码验证结果，正在重试")
                continue
            geetest_result = await geetest_up.text_content()
            assert geetest_result
            logger.debug(f"[Captcha] Geetest result: {geetest_result}")
            if "验证成功" in geetest_result:
                logger.success("[Captcha] 极验网页 Tip 验证成功")
                captcha_image_body = ""
                await page.wait_for_timeout(2000)
            else:
                logger.warning("[Captcha] 极验验证失败，正在重试")

    return page
